#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>

#define LOOPS    300
#define WARMUP   10

#define MAX_LEN (8*1024*1024)	/* 8 MB */

static unsigned char *main_buffer = NULL;

/* fill the buffer with 'random' caracters */
void fill_buffer(char *buffer, int len)
{
  int i;
  for(i=0;i<len; i++) {
    buffer[i]='a'+(i%26);
  }
}

int
main(int    argc,
     char **argv)
{
  int	comm_rank	= -1;
  int	comm_size	= -1;
  char	host_name[1024]	= "";
  int dest;
  int len;
  int iterations     = LOOPS;

  MPI_Init(&argc,&argv);

  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);

  if (gethostname(host_name, 1023) < 0) {
    perror("gethostname");
    exit(1);
  }

  if (comm_size != 2) {
    fprintf(stderr, "This program requires 2 MPI processes, aborting...\n");
    goto out;
  }

  dest = (comm_rank+1)%2;

  if(! comm_rank) {
    printf("#iter\tsize\tlatency\tbandwidth\n");
  }

  /* measure the average transfer duration for various message lengths */
  for(len = 1; len <= MAX_LEN; len*=2) {
    main_buffer = malloc(len);
    fill_buffer(main_buffer, len);

    if(len > (32*1024)) {
      /* reduce the number of iterations for large messages */
      iterations = iterations / 2;
    }
    int i;

    MPI_Barrier(MPI_COMM_WORLD);
    /* a few warmup iterations  */
    for(i = 0; i< WARMUP; i++) {
      MPI_Send(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD);
      MPI_Recv(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }

    double t1, t2;
    /* measurement */
    t1 = MPI_Wtime();
    for(i = 0; i< iterations; i++) {
      MPI_Send(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD);
      MPI_Recv(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }
    t2 = MPI_Wtime();

    if(! comm_rank) {
      double latency = 1e6*(t2-t1)/(2*iterations);
      double bandwidth = 2*len/latency;
      printf("%d\t%d\t%lf\n",iterations, len, bandwidth);
    }
  }

 out:
  free(main_buffer);
  MPI_Finalize();

  return 0;
}
