#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <mpi.h>

#define LOOPS    10
#define WARMUP   1

#define MAX_LEN (8*1024*1024)	/* 8 MB */

static unsigned char *main_buffer = NULL;

/* fill the buffer with 'random' caracters */
void fill_buffer(char *buffer, int len)
{
  int i;
  for(i=0;i<len; i++) {
    buffer[i]='a'+(i%26);
  }
}

/* simulate a computation */
void compute(unsigned usec)
{
  double t1, t2;
  t1 = MPI_Wtime();
  do{
    t2 = MPI_Wtime();
  } while ( (t2-t1)*1e6 <usec);
}

int
main(int    argc,
     char **argv)
{
  int	comm_rank	= -1;
  int	comm_size	= -1;
  char	host_name[1024]	= "";
  int dest;
  int len;
  int iterations     = LOOPS;

  MPI_Init(&argc,&argv);

  MPI_Comm_size(MPI_COMM_WORLD, &comm_size);
  MPI_Comm_rank(MPI_COMM_WORLD, &comm_rank);

  if (gethostname(host_name, 1023) < 0) {
    perror("gethostname");
    exit(1);
  }

  if (comm_size != 2) {
    fprintf(stderr, "This program requires 2 MPI processes, aborting...\n");
    goto out;
  }

  dest = (comm_rank+1)%2;

  if(! comm_rank) {
    printf("#iter\tsize\tlatency\tbandwidth\n");
  }

  /* measure the average transfer duration for various message lengths */
  for(len = 1; len <= MAX_LEN; len*=2) {
    main_buffer = malloc(len);
    fill_buffer(main_buffer, len);

    int i;

    /* warm up so that the measurement is accurate */
    for(i = 0; i< WARMUP; i++) {
      MPI_Request req;
      if(! comm_rank) {
	MPI_Send(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD);

	MPI_Irecv(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD, &req);
	MPI_Wait(&req, MPI_STATUS_IGNORE);
      } else {
	MPI_Irecv(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD, &req);
	MPI_Wait(&req, MPI_STATUS_IGNORE);

	MPI_Send(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD);
      }
    }

    /* measurement */
    double t1, t2, total_time=0;

    MPI_Barrier(MPI_COMM_WORLD);
    for(i = 0; i< iterations; i++) {
      MPI_Request req;
      if(! comm_rank) {
	t1 = MPI_Wtime();
	MPI_Send(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	total_time += t2-t1;

	MPI_Irecv(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD, &req);
	MPI_Wait(&req, MPI_STATUS_IGNORE);
      } else {
	MPI_Irecv(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD, &req);
	compute(100000);
	MPI_Wait(&req, MPI_STATUS_IGNORE);

	MPI_Send(main_buffer, len, MPI_CHAR, dest, 0, MPI_COMM_WORLD);
      }
    }
    MPI_Barrier(MPI_COMM_WORLD);

    if(! comm_rank) {
      double latency = 1e6*(total_time)/(2*iterations);
      printf("%d\t%d\t%lf\n",iterations, len, latency);
    }
  }

 out:
  free(main_buffer);
  MPI_Finalize();

  return 0;
}
